@@ -6,6 +6,7 @@
 ------------------------------------------------------------------------------*/
 #include "CLucene/StdHeader.h"
 #include "IndexReader.h"
+#include "IndexWriter.h"
 
 #include "CLucene/store/Directory.h"
 #include "CLucene/store/FSDirectory.h"
@@ -92,12 +93,12 @@ CL_NS_DEF(index)
 
 	  LuceneLock* lock = directory->makeLock("commit.lock");
 
-	  //Instantiate an IndexReaderLockWith which can produce an IndexReader
-      IndexReaderLockWith with(lock,directory);
+	  //Instantiate an IndexReader::LockWith which can produce an IndexReader
+      IndexReader::LockWith with(lock,directory);
 
 	  try{
 	  //Create an IndexReader reading the index
-		ret = (IndexReader*)with.run();
+		ret = with.runAndReturn();
 	  }_CLFINALLY(
         _CLDELETE( lock );
 	  );
@@ -108,8 +109,15 @@ CL_NS_DEF(index)
 	   //return reference 
        return ret;
   }
+  
+  CL_NS(document)::Document* IndexReader::document(const int32_t n){
+    CL_NS(document)::Document* ret = _CLNEW CL_NS(document)::Document;
+    if (!document(n,ret) )
+        _CLDELETE(ret);
+    return ret;
+  }
 
-  void* IndexReader::IndexReaderLockWith::doBody() {
+  IndexReader* IndexReader::LockWith::doBody() {
   //Func - Reads the segmentinfo file and depending on the number of segments found
   //       it returns a SegmentsReader or a SegmentReader
   //Pre  - directory != NULL
@@ -151,7 +159,7 @@ CL_NS_DEF(index)
 			if (nSegs > 0){
 				uint32_t infosize=infos->size();
 				readers = _CL_NEWARRAY(IndexReader*,infosize+1);
-				for (uint32_t i = 0; i < infosize; i++) {
+				for (uint32_t i = 0; i < infosize; ++i) {
 					//Instantiate a SegementReader responsible for reading the i-th segment and store it in
 					//the readers array
 					readers[i] = _CLNEW SegmentReader(infos->info(i));
@@ -176,7 +184,19 @@ CL_NS_DEF(index)
   }
 
   int64_t IndexReader::getCurrentVersion(Directory* directory) {
-     return SegmentInfos::readCurrentVersion(directory);
+	SCOPED_LOCK_MUTEX(directory->THIS_LOCK)                 // in- & inter-process sync
+	LuceneLock* commitLock=directory->makeLock(IndexWriter::COMMIT_LOCK_NAME);
+	bool locked=false;
+	int64_t ret = 0;
+	try {
+		locked=commitLock->obtain(IndexWriter::COMMIT_LOCK_TIMEOUT);
+		ret = SegmentInfos::readCurrentVersion(directory);
+	}_CLFINALLY(
+		if (locked) {
+			commitLock->release();
+		}
+	)
+	return ret;
   }
 
 
@@ -187,7 +207,26 @@ CL_NS_DEF(index)
       _CLDECDELETE(dir);
       return version;
    }
-    
+    int64_t IndexReader::getVersion() {
+		return segmentInfos->getVersion();
+	}
+	
+	bool IndexReader::isCurrent() {
+		SCOPED_LOCK_MUTEX(directory->THIS_LOCK)                 // in- & inter-process sync
+		LuceneLock* commitLock = directory->makeLock(IndexWriter::COMMIT_LOCK_NAME);
+		bool locked=false;
+		bool ret = false;
+		try {
+			locked=commitLock->obtain(IndexWriter::COMMIT_LOCK_TIMEOUT);
+			ret = SegmentInfos::readCurrentVersion(directory) == segmentInfos->getVersion();
+		} _CLFINALLY(
+			if (locked) {
+				commitLock->release();
+			}
+		)
+		return ret;
+	}
+
   uint64_t IndexReader::lastModified(const Directory* directory) {
   //Func - Static method
   //       Returns the time the index in this directory was last modified. 
@@ -209,9 +248,9 @@ CL_NS_DEF(index)
        CND_PRECONDITION(directory != NULL, "directory is NULL");
 
 	   //Create a buffer of length CL_MAXDIR
-       char f[CL_MAX_PATH+10]; //add 10 in case that directory is already 260 long
-	   //Copy the directory string to the buffer
-       strcpy(f,directory);
+       char f[CL_MAX_PATH];
+	   //Copy the directory string to the buffer. leave room for /segments
+       strncpy(f,directory,CL_MAX_PATH-10);
 	   //Cat the name of the segments to buffer
        strcat(f, "/segments");
 	   //Check if the segments file exists
@@ -233,7 +272,7 @@ CL_NS_DEF(index)
 
     if (writeLock == NULL) {
       LuceneLock* writeLock = directory->makeLock("write.lock");
-      if (!writeLock->obtain(LUCENE_WRITE_LOCK_TIMEOUT)) // obtain write lock
+      if (!writeLock->obtain(IndexWriter::WRITE_LOCK_TIMEOUT)) // obtain write lock
        _CLTHROWA(CL_ERR_IO,"Index locked for write"); // + writeLock
       this->writeLock = writeLock;
 
@@ -346,7 +385,7 @@ CL_NS_DEF(index)
 	        SCOPED_LOCK_MUTEX(directory->THIS_LOCK)      // in- & inter-process sync
 	
 	        LuceneLock* commitLock = directory->makeLock("commit.lock");
-	        IndexReaderCommitLockWith cl(commitLock,this);
+	        IndexReader::CommitLockWith cl(commitLock,this);
 	        cl.run();
 			_CLDELETE(commitLock);
 	
@@ -396,7 +435,7 @@ CL_NS_DEF(index)
           while (docs->next()) {
 			  //Delete the document
               deleteDocument(docs->doc());
-              Counter++;
+              ++Counter;
           }
       }_CLFINALLY(
 		  //Close the enumeration
@@ -404,11 +443,43 @@ CL_NS_DEF(index)
           );
 
     //Delete the enumeration of found documents
-    _CLVDELETE( docs ); //todo: not a clucene object... should be
+    _CLDELETE( docs );
 
 	//Return the number of deleted documents
     return Counter;
   }
+  
+	TCHAR** IndexReader::getFieldNames(){
+		CL_NS(util)::StringArrayWithDeletor array;
+		getFieldNames(IndexReader::ALL, array);
+		
+		array.setDoDelete(false);
+		TCHAR** ret = _CL_NEWARRAY(TCHAR*,array.size()+1);
+		int j=0;
+  		CL_NS(util)::StringArrayWithDeletor::iterator itr = array.begin();
+  		while ( itr != array.end() ){
+  			ret[j]=*itr;
+			++j;++itr;
+ 		}
+		ret[j]=NULL;
+		return ret;
+	}
+	TCHAR** IndexReader::getFieldNames(bool indexed){
+		CL_NS(util)::StringArrayWithDeletor array;
+		getFieldNames(indexed?IndexReader::INDEXED:IndexReader::UNINDEXED, array);
+		
+		array.setDoDelete(false);
+		TCHAR** ret = _CL_NEWARRAY(TCHAR*,array.size()+1);
+		int j=0;
+  		CL_NS(util)::StringArrayWithDeletor::iterator itr = array.begin();
+  		while ( itr != array.end() ){
+  			ret[j]=*itr;
+			++j;++itr;
+ 		}
+		ret[j]=NULL;
+		return ret;
+	}
+  
 
   void IndexReader::close() {
   //Func - Closes files associated with this index and also saves any new deletions to disk.
@@ -460,9 +531,9 @@ CL_NS_DEF(index)
       CND_PRECONDITION(directory != NULL, "directory is NULL");
 
 	  //Create a buffer of length CL_MAXDIR
-      char f[CL_MAX_PATH+12]; //add 12 in case that directory is already 260 long
-	  //Copy the directory string to the buffer
-      strcpy(f,directory);
+      char f[CL_MAX_PATH]; //todo: potential buffer overflow
+	  //Copy the directory string to the buffer. leave room for /write.lock
+      strncpy(f,directory,CL_MAX_PATH-12);
 	  //Cat the name of the write.lock file to buffer
       strcat ( f,"/write.lock" );
 
@@ -473,13 +544,20 @@ CL_NS_DEF(index)
 
 	  return ret;
   }
+  
+/** Returns true if there are norms stored for this field. */
+bool IndexReader::hasNorms(const TCHAR* field) {
+	// backward compatible implementation.
+	// SegmentReader has an efficient implementation.
+	return norms(field) != NULL;
+}
 
-  void IndexReader::unlock(const char* path){
-	  FSDirectory* dir = FSDirectory::getDirectory(path,false);
-	  unlock(dir);
-	  dir->close();
-	  _CLDECDELETE(dir);
-  }
+void IndexReader::unlock(const char* path){
+	FSDirectory* dir = FSDirectory::getDirectory(path,false);
+	unlock(dir);
+	dir->close();
+	_CLDECDELETE(dir);
+}
   void IndexReader::unlock(Directory* directory){
   //Func - Static method
   //       Forcibly unlocks the index in the named directory->
@@ -499,27 +577,79 @@ CL_NS_DEF(index)
       _CLDELETE(lock);
   }
 
+bool IndexReader::isLuceneFile(const char* filename){
+	if ( !filename )
+		return false;
+	size_t len = strlen(filename);
+	if ( len < 6 ) //need at least x.frx
+		return false;
+	const char* ext=filename+len;
+	while(*ext != '.' && ext!=filename)
+		ext--;
+
+	if ( strcmp(ext, ".cfs") == 0 )
+		return true;
+	else if ( strcmp(ext, ".fnm") == 0 )
+		return true;
+	else if ( strcmp(ext, ".fdx") == 0 )
+		return true;
+	else if ( strcmp(ext, ".fdt") == 0 )
+		return true;
+	else if ( strcmp(ext, ".tii") == 0 )
+		return true;
+	else if ( strcmp(ext, ".tis") == 0 )
+		return true;
+	else if ( strcmp(ext, ".frq") == 0 )
+		return true;
+	else if ( strcmp(ext, ".prx") == 0 )
+		return true;
+	else if ( strcmp(ext, ".del") == 0 )
+		return true;
+	else if ( strcmp(ext, ".tvx") == 0 )
+		return true;
+	else if ( strcmp(ext, ".tvd") == 0 )
+		return true;
+	else if ( strcmp(ext, ".tvf") == 0 )
+		return true;
+	else if ( strcmp(ext, ".tvp") == 0 )
+		return true;
+
+	else if ( strcmp(filename, "segments") == 0 )
+		return true;
+	else if ( strcmp(filename, "segments.new") == 0 )
+		return true;
+	else if ( strcmp(filename, "deletable") == 0 )
+		return true;
+
+	else if ( strncmp(ext,".f",2)==0 ){
+		const char* n = ext+2;
+		if ( *n && _istdigit(*n) )
+			return true;	
+	}
+
+	return false;
+}
+
 	void IndexReader::addCloseCallback(CloseCallback callback, void* parameter){
 		closeCallbacks.put(callback, parameter);	
 	}
 
 
 	//Constructor	
-    IndexReader::IndexReaderLockWith::IndexReaderLockWith(CL_NS(store)::LuceneLock* lock, CL_NS(store)::Directory* dir):
-		CL_NS(store)::LuceneLockWith(lock,LUCENE_COMMIT_LOCK_TIMEOUT)
+    IndexReader::LockWith::LockWith(CL_NS(store)::LuceneLock* lock, CL_NS(store)::Directory* dir):
+		CL_NS(store)::LuceneLockWith<IndexReader*>(lock,IndexWriter::COMMIT_LOCK_TIMEOUT)
 	{
 		this->directory = dir;
 	}	
 	//Constructor	
-	IndexReader::IndexReaderCommitLockWith::IndexReaderCommitLockWith( CL_NS(store)::LuceneLock* lock, IndexReader* r ):
-		CL_NS(store)::LuceneLockWith(lock,LUCENE_COMMIT_LOCK_TIMEOUT),
+	IndexReader::CommitLockWith::CommitLockWith( CL_NS(store)::LuceneLock* lock, IndexReader* r ):
+		CL_NS(store)::LuceneLockWith<void>(lock,IndexWriter::COMMIT_LOCK_TIMEOUT),
 		reader(r)
 	{
 	}
-	void* IndexReader::IndexReaderCommitLockWith::doBody(){
+	void IndexReader::CommitLockWith::doBody(){
 		reader->doCommit();
 		reader->segmentInfos->write(reader->getDirectory());
-		return NULL;
 	}
 
 CL_NS_END
