@@ -12,19 +12,26 @@
 #include "IndexReader.h"
 #include "TermInfosReader.h"
 #include "Terms.h"
+#include "CLucene/search/Similarity.h"
 
 CL_NS_USE(util)
 CL_NS_USE(store)
 CL_NS_USE(document)
+CL_NS_USE(search)
 CL_NS_DEF(index)
 
  SegmentReader::Norm::Norm(IndexInput* instrm, int32_t n, SegmentReader* r, const char* seg): 
-   in(instrm), number(n), reader(r), segment(seg){
+	number(n), 
+	reader(r), 
+	segment(seg), 
+	in(instrm),
+	bytes(NULL), 
+	dirty(false){
   //Func - Constructor
   //Pre  - instrm is a valid reference to an IndexInput
   //Post - A Norm instance has been created with an empty bytes array
 
-	  bytes = NULL;
+	 bytes = NULL;
      dirty = false;
   }
 
@@ -54,7 +61,10 @@ CL_NS_DEF(index)
         out->writeBytes(bytes, reader->maxDoc());
       }_CLFINALLY( out->close(); _CLDELETE(out) );
 
-      sprintf(fileName,"%s.f%d",segment,number);
+	  if ( reader->cfsReader == NULL )
+		sprintf(fileName,"%s.f%d",segment,number);
+	  else // use a different file name if we have compound format
+		sprintf(fileName,"%s.s%d",segment,number);
       reader->getDirectory()->renameFile(buf, fileName);
       this->dirty = false;
     }
@@ -81,6 +91,7 @@ CL_NS_DEF(index)
       //Post - All files of the segment have been read
 
       deletedDocs      = NULL;
+	  ones			   = NULL;
 	  //There are no documents yet marked as deleted
       deletedDocsDirty = false;
       
@@ -118,9 +129,11 @@ CL_NS_DEF(index)
 
 	  //Open an IndexInput freqStream to the frequency file
 #ifdef LUCENE_FS_MMAP
-	  if ( strcmp(cfsDir->getDirectoryType(), "FS") == 0 ){
+	  if ( cfsDir->getDirectoryType() == FSDirectory::DirectoryType() ){
 		  FSDirectory* fsdir = (FSDirectory*)cfsDir;
 		  freqStream = fsdir->openMMapFile( buf );
+      } else if (strcmp(cfsDir->getDirectoryType(), "CFS") == 0) { //todo: we should have a CFS Directory
+          freqStream = cfsDir->openInput(buf,true);
 	  }else
 #endif
 		freqStream = cfsDir->openInput( buf );
@@ -133,6 +146,14 @@ CL_NS_DEF(index)
       SegmentName(buf, CL_MAX_PATH,".prx");
 
 	  //Open an IndexInput proxStream to the prox file
+#ifdef LUCENE_FS_MMAP
+    if (cfsDir->getDirectoryType() == FSDirectory::DirectoryType()) {
+        FSDirectory* fsdir = (FSDirectory*)cfsDir;
+        proxStream = fsdir->openMMapFile( buf );
+    } else if (strcmp(cfsDir->getDirectoryType(), "CFS") == 0) {
+        proxStream = cfsDir->openInput(buf,true);
+    } else
+#endif
       proxStream = cfsDir->openInput( buf );
 
 	  //Condition check to see if proxStream points to a valid instance and was able to open the
@@ -166,9 +187,9 @@ CL_NS_DEF(index)
       openNorms(cfsDir);
 
       if (fieldInfos->hasVectors()) { // open term vector files only as needed
-         termVectorsReader = _CLNEW TermVectorsReader(cfsDir, segment, fieldInfos);
+         termVectorsReaderOrig = _CLNEW TermVectorsReader(cfsDir, segment, fieldInfos);
       }else
-		  termVectorsReader = NULL;
+		 termVectorsReaderOrig = NULL;
   }
 
   SegmentReader::~SegmentReader(){
@@ -185,8 +206,10 @@ CL_NS_DEF(index)
 	  _CLDELETE(proxStream);
 	  _CLDELETE_CaARRAY(segment);
 	  _CLDELETE(deletedDocs);
-     _CLDELETE(termVectorsReader)
+	  _CLDELETE_ARRAY(ones);
+     _CLDELETE(termVectorsReaderOrig)
      _CLDECDELETE(cfsReader);
+    //termVectorsLocal->unregister(this);
   }
 
   void SegmentReader::doCommit(){
@@ -195,20 +218,19 @@ CL_NS_DEF(index)
    strcat(bufdel,".del");
 
     if (deletedDocsDirty) {               // re-write deleted 
-       char buftmp[CL_MAX_PATH];
-       strcpy(buftmp,segment);
-       strcat(buftmp,".tmp");
-      deletedDocs->write(getDirectory(), buftmp);
-      getDirectory()->renameFile(buftmp,bufdel);
+		char buftmp[CL_MAX_PATH];
+		strcpy(buftmp,segment);
+		strcat(buftmp,".tmp");
+		deletedDocs->write(getDirectory(), buftmp);
+		getDirectory()->renameFile(buftmp,bufdel);
     }
     if(undeleteAll && getDirectory()->fileExists(bufdel)){
-      getDirectory()->deleteFile(bufdel);
+      getDirectory()->deleteFile(bufdel, true);
     }
     if (normsDirty) {               // re-write norms 
-	  CL_NS(util)::CLHashtable<const TCHAR*,Norm*,Compare::TChar,Equals::TChar>::iterator itr = _norms.begin();
-		  Norm* norm;
+	  NormsType::iterator itr = _norms.begin();
       while (itr != _norms.end()) {
-        norm = itr->second;
+        Norm* norm = itr->second;
         if (norm->dirty) {
           norm->reWrite();
         }
@@ -246,14 +268,14 @@ CL_NS_DEF(index)
 	  //Close the norm file
       closeNorms();
     
-     if (termVectorsReader != NULL) 
-        termVectorsReader->close();
+     if (termVectorsReaderOrig != NULL) 
+        termVectorsReaderOrig->close();
 
      if (cfsReader != NULL)
          cfsReader->close();
   }
 
-  bool SegmentReader::hasDeletions() {
+  bool SegmentReader::hasDeletions()  const{
       return deletedDocs != NULL;
   }
 
@@ -267,27 +289,25 @@ CL_NS_DEF(index)
   
   //static
   bool SegmentReader::hasSeparateNorms(SegmentInfo* si) {
-    char** result = si->getDir()->list();
+  	vector<string> names;
+  	si->getDir()->list(&names);
+  	
     char pattern[CL_MAX_PATH];
     strcpy(pattern,si->name);
-    strcat(pattern,".f");
+    strcat(pattern,".s");
     size_t patternLength = strlen(pattern);
 
-    int32_t i=0;
-    char* res=NULL;
+    string res;
 	bool ret=false;
-    while ( (res=result[i]) != NULL ){
-		if ( !ret ){
-			if ( strlen(res)>patternLength && strncmp(res,pattern,patternLength) == 0 ){
-				if ( res[patternLength] >= '0' && res[patternLength] <= '9' )
-					ret=true;
-			}
+	vector<string>::iterator itr = names.begin();
+    while ( itr != names.end() ){
+		if ( (*itr).length()>patternLength && strncmp((*itr).c_str(),pattern,patternLength) == 0 ){
+			if ( (*itr).at(patternLength) >= '0' && (*itr).at(patternLength) <= '9' )
+				return true;
 		}
-	  _CLDELETE_CaARRAY(result[i]);
-	  ++i;
+		itr++;
     }
-	_CLDELETE_ARRAY(result);
-    return ret;
+    return false;
   }
 
   bool SegmentReader::hasDeletions(const SegmentInfo* si) {
@@ -339,22 +359,16 @@ CL_NS_DEF(index)
       undeleteAll = true;
   }
 
-  AStringArrayConstWithDeletor* SegmentReader::files() {
+  void SegmentReader::files(CL_NS(util)::AStringArrayWithDeletor& retarray) {
   //Func - Returns all file names managed by this SegmentReader
   //Pre  - segment != NULL
   //Post - All filenames managed by this SegmentRead have been returned
  
      CND_PRECONDITION(segment != NULL, "segment is NULL");
 
-	  AStringArrayConstWithDeletor* files = _CLNEW AStringArrayConstWithDeletor(true);
-
-     //Condition check to see if files points to a valid instance
-     CND_CONDITION(files != NULL, "No memory could be allocated for files");
-
      char* temp = NULL;
-     #define _ADD_SEGMENT(ext) temp = SegmentName( ext ); if ( getDirectory()->fileExists(temp) ) files->push_back(temp); else _CLDELETE_CaARRAY(temp);
+     #define _ADD_SEGMENT(ext) temp = SegmentName( ext ); if ( getDirectory()->fileExists(temp) ) retarray.push_back(temp); else _CLDELETE_CaARRAY(temp);
 								
-
      //Add the name of the Field Info file
      _ADD_SEGMENT(".cfs" );
      _ADD_SEGMENT(".fnm" );
@@ -376,13 +390,19 @@ CL_NS_DEF(index)
           //Get the field info for the i-th field   
           fi = fieldInfos->fieldInfo(i);
           //Check if the field has been indexed
-          if (fi->isIndexed){
+          if (fi->isIndexed && !fi->omitNorms){
+			  char* name;
+			  if ( cfsReader == NULL )
+				  name = SegmentName(".f", i);
+			  else
+				  name = SegmentName(".s", i);
               //The field has been indexed so add its norm file
-              files->push_back( SegmentName(".f", i) );
+              if ( getDirectory()->fileExists(name) )
+				  retarray.push_back( name );
+			  else
+				  _CLDELETE_CaARRAY(name);
           }
        }
-
-    return files;
   }
 
   TermEnum* SegmentReader::terms() const {
@@ -407,7 +427,7 @@ CL_NS_DEF(index)
       return tis->terms(t);
   }
 
-  Document* SegmentReader::document(const int32_t n) {
+  bool SegmentReader::document(int32_t n, Document* doc) {
   //Func - Returns a document identified by n
   //Pre  - n >=0 and identifies the document n
   //Post - if the document has been deleted then an exception has been thrown
@@ -420,20 +440,14 @@ CL_NS_DEF(index)
 	  //Check if the n-th document has been marked deleted
        if (isDeleted(n)){
           _CLTHROWA( CL_ERR_InvalidState,"attempt to access a deleted document" );
-          }
+       }
 
 	   //Retrieve the n-th document
-       Document* ret = fieldsReader->doc(n);
-
-       //Condition check to see if ret points to a valid instance
-       CND_CONDITION(ret != NULL, "No document could be retrieved");
-
-	   //Return the document
-       return ret;
+       return fieldsReader->doc(n, doc);
   }
 
 
-  bool SegmentReader::isDeleted(const int32_t n) {
+  bool SegmentReader::isDeleted(const int32_t n){
   //Func - Checks if the n-th document has been marked deleted
   //Pre  - n >=0 and identifies the document n
   //Post - true has been returned if document n has been deleted otherwise fralse
@@ -512,6 +526,46 @@ CL_NS_DEF(index)
       return fieldsReader->size();
   }
 
+void SegmentReader::getFieldNames(FieldOption fldOption, StringArrayWithDeletor& retarray){
+	size_t len = fieldInfos->size();
+	for (size_t i = 0; i < len; i++) {
+		FieldInfo* fi = fieldInfos->fieldInfo(i);
+		bool v=false;
+		if (fldOption & IndexReader::ALL) {
+			v=true;
+		}else {
+			if (!fi->isIndexed && (fldOption & IndexReader::UNINDEXED) )
+				v=true;
+			if (fi->isIndexed && (fldOption & IndexReader::INDEXED) )
+				v=true;
+			if (fi->isIndexed && fi->storeTermVector == false && ( fldOption & IndexReader::INDEXED_NO_TERMVECTOR) )
+				v=true;
+			if ( (fldOption & IndexReader::TERMVECTOR) &&
+				    fi->storeTermVector == true &&
+					fi->storePositionWithTermVector == false &&
+					fi->storeOffsetWithTermVector == false )
+				v=true;
+			if (fi->isIndexed && fi->storeTermVector && (fldOption & IndexReader::INDEXED_WITH_TERMVECTOR) )
+				v=true;
+			if (fi->storePositionWithTermVector && fi->storeOffsetWithTermVector == false && 
+					(fldOption & IndexReader::TERMVECTOR_WITH_POSITION))
+				v=true;
+			if (fi->storeOffsetWithTermVector && fi->storePositionWithTermVector == false && 
+					(fldOption & IndexReader::TERMVECTOR_WITH_OFFSET) )
+				v=true;
+			if ((fi->storeOffsetWithTermVector && fi->storePositionWithTermVector) &&
+					(fldOption & IndexReader::TERMVECTOR_WITH_POSITION_OFFSET) )
+				v=true;
+		}
+		if ( v )
+			retarray.push_back(STRDUP_TtoT(fi->name));
+	}
+}
+
+bool SegmentReader::hasNorms(const TCHAR* field) const{
+	return _norms.find(field) != _norms.end();
+}
+
 
   void SegmentReader::norms(const TCHAR* field, uint8_t* bytes) {
   //Func - Reads the Norms for field from disk starting at offset in the inputstream
@@ -524,14 +578,15 @@ CL_NS_DEF(index)
     CND_PRECONDITION(field != NULL, "field is NULL");
     CND_PRECONDITION(bytes != NULL, "field is NULL");
 
-	 SCOPED_LOCK_MUTEX(THIS_LOCK)
+	SCOPED_LOCK_MUTEX(THIS_LOCK)
     
     Norm* norm = _norms.get(field);
-    if ( norm == NULL )
-       return;					  // use zeros in array
-
-   if (norm->bytes != NULL) { // can copy from cache
-      memcpy(bytes,norm->bytes,maxDoc());
+	if ( norm == NULL ){
+		memcpy(bytes, fakeNorms(), maxDoc());
+		return;
+	}
+    if (norm->bytes != NULL) { // can copy from cache
+	  memcpy(bytes, norm->bytes, maxDoc());
       return;
     }
 
@@ -547,8 +602,33 @@ CL_NS_DEF(index)
         _normStream->close();
         //Destroy the normstream
         _CLDELETE( _normStream );
-	);
-	
+	);	
+  }
+
+  uint8_t* SegmentReader::createFakeNorms(int32_t size) {
+    uint8_t* ones = _CL_NEWARRAY(uint8_t,size);
+	memset(ones, DefaultSimilarity::encodeNorm(1.0f), size);
+    return ones;
+  }
+
+  uint8_t* SegmentReader::fakeNorms() {
+    if (ones==NULL) 
+		ones=createFakeNorms(maxDoc());
+    return ones;
+  }
+  // can return null if norms aren't stored
+  uint8_t* SegmentReader::getNorms(const TCHAR* field) {
+	SCOPED_LOCK_MUTEX(THIS_LOCK)
+    Norm* norm = _norms.get(field);
+    if (norm == NULL) 
+		return NULL;  // not indexed, or norms not stored
+
+    if (norm->bytes == NULL) {                     // value not yet read
+      uint8_t* bytes = _CL_NEWARRAY(uint8_t, maxDoc());
+      norms(field, bytes);
+      norm->bytes = bytes;                         // cache it
+    }
+    return norm->bytes;
   }
 
   uint8_t* SegmentReader::norms(const TCHAR* field) {
@@ -559,32 +639,11 @@ CL_NS_DEF(index)
   //       and returned containing the norms for that field. If the named field is unknown NULL is returned.
 
     CND_PRECONDITION(field != NULL, "field is NULL");
-    
-    
-	  SCOPED_LOCK_MUTEX(THIS_LOCK)
-
-    //Try to retrieve the norms for field
-    Norm* norm = (Norm*)_norms.get(field);
-    //Check if a norm instance was found
-    if (norm == NULL){
-        //return NULL as there are no norms to be returned
-        return NULL;
-	}
-
-    if (norm->bytes == NULL) { //value not read yet
-        //allocate a new bytes array to hold the norms
-        uint8_t* bytes = _CL_NEWARRAY(uint8_t,maxDoc()); 
-
-        //Condition check to see if bytes points to a valid array
-        CND_CONDITION(bytes != NULL, "bytes is NULL");
-
-        //Read the norms from disk straight into the new bytes array
-        norms(field, bytes);
-        norm->bytes = bytes; // cache it
-    }
-
-		//Return the norms
-		return norm->bytes;
+    SCOPED_LOCK_MUTEX(THIS_LOCK)
+	uint8_t* bytes = getNorms(field);
+    if (bytes==NULL) 
+		bytes=fakeNorms();
+    return bytes;
   }
 
   void SegmentReader::doSetNorm(int32_t doc, const TCHAR* field, uint8_t value){
@@ -639,25 +698,23 @@ CL_NS_DEF(index)
   //       a corresponding norm file ready to be read
 
       //Iterate through all the fields
-			FieldInfo* fi;
-			Directory* d;
-      for (int32_t i = 0; i < fieldInfos->size(); ++i) {
+      for (int32_t i = 0; i < fieldInfos->size(); i++) {
 		  //Get the FieldInfo for the i-th field
-          fi = fieldInfos->fieldInfo(i);
+          FieldInfo* fi = fieldInfos->fieldInfo(i);
           //Check if the field is indexed
-          if (fi->isIndexed) {
+		  if (fi->isIndexed && !fi->omitNorms ) {
 		      //Allocate a buffer
               char fileName[CL_MAX_PATH];
-			  //Create a filename for the norm file
-              SegmentName(fileName,CL_MAX_PATH, ".f", fi->number);
-              //TODO, should fi->name be copied?
-			  //Create a new Norm with an open inputstream to f and store
-			  //it at fi->name in norms
-             d = getDirectory();
-             if ( !d->fileExists(fileName) )
-                d = cfsDir;
-
-                 _norms.put(fi->name, _CLNEW Norm( d->openInput( fileName ),fi->number, this, segment ));
+			  
+			  // look first if there are separate norms in compound format
+              SegmentName(fileName,CL_MAX_PATH, ".s", fi->number);
+			  Directory* d = getDirectory();
+			  if(!d->fileExists(fileName)){
+				SegmentName(fileName,CL_MAX_PATH, ".f", fi->number);
+				d = cfsDir;
+			  }
+             
+			  _norms.put(fi->name, _CLNEW Norm( d->openInput( fileName ),fi->number, this, segment ));
           }
       }
   }
@@ -669,7 +726,7 @@ CL_NS_DEF(index)
 
     SCOPED_LOCK_MUTEX(_norms.THIS_LOCK)
 	//Create an interator initialized at the beginning of norms
-	CL_NS(util)::CLHashtable<const TCHAR*,Norm*,Compare::TChar,Equals::TChar>::iterator itr = _norms.begin();
+	NormsType::iterator itr = _norms.begin();
 	//Iterate through all the norms
     while (itr != _norms.end()) {
         //Get the norm
@@ -683,105 +740,38 @@ CL_NS_DEF(index)
     _norms.clear(); //bvk: they're deleted, so clear them so that they are not re-used
   }
 
-    /**
-   * @see IndexReader#getFieldNames()
-   */
-  TCHAR** SegmentReader::getFieldNames(){
-    // maintain a unique set of field names
-	int32_t len = fieldInfos->size();
-    TCHAR** ret = _CL_NEWARRAY(TCHAR*,len+1);
-	int32_t i = 0;
-	int32_t p = 0;
-	FieldInfo* fi;
-	int32_t j;
-    for (i = 0; i < len; ++i) {
-      fi = fieldInfos->fieldInfo(i);
-	   for ( j =0;j<i;++j )
-		  if ( _tcscmp(fi->name,ret[j]) == 0 )
-			  continue;
-      ret[p++]=STRDUP_TtoT(fi->name);
-    }
-    ret[p]=NULL;
-    return ret;
-  }
 
-  /**
-   * @see IndexReader#getFieldNames(boolean)
-   */
-  TCHAR** SegmentReader::getFieldNames(bool indexed) {
-    // maintain a unique set of field names
-    CL_NS(util)::CLSetList<const TCHAR*> fieldSet(false);
-	int32_t i = 0;
-	FieldInfo* fi;
-    for (i = 0; i < fieldInfos->size(); ++i) {
-      fi = fieldInfos->fieldInfo(i);
-	   if (fi->isIndexed == indexed){
-		 if ( fieldSet.find(fi->name)==fieldSet.end() )
-			  fieldSet.insert(fi->name);
-	   }
-    }
-    
-	 TCHAR** ret = _CL_NEWARRAY(TCHAR*,fieldSet.size()+1);
-    int j=0;
-    CL_NS(util)::CLSetList<const TCHAR*>::iterator itr = fieldSet.begin();
-    while ( itr != fieldSet.end() ){
-        const TCHAR* t = *itr;
-        ret[j]=STRDUP_TtoT(t);
-
-        ++j;
-        ++itr;
-    }
-    ret[fieldSet.size()]=NULL;
-    return ret;
-  }
-
-  /**
-   * 
-   * @param storedTermVector if true, returns only Indexed fields that have term vector info, 
-   *                        else only indexed fields without term vector info 
-   * @return Collection of Strings indicating the names of the fields
-   */
-   TCHAR** SegmentReader::getIndexedFieldNames(bool storedTermVector) {
-    // maintain a unique set of field names
-    CL_NS(util)::CLSetList<const TCHAR*> fieldSet(false);
-	 int32_t i = 0;
-	  FieldInfo* fi;
-    for (i = 0; i < fieldInfos->size(); ++i) {
-      fi = fieldInfos->fieldInfo(i);
-      if (fi->isIndexed == true && fi->storeTermVector == storedTermVector){
-		if ( fieldSet.find((const TCHAR*)fi->name)==fieldSet.end() )
-			  fieldSet.insert(fi->name);
-      }
-    }
-	 TCHAR** ret = _CL_NEWARRAY(TCHAR*,fieldSet.size()+1);
-    int j=0;
-    CL_NS(util)::CLSetList<const TCHAR*>::iterator itr = fieldSet.begin();
-    while ( itr != fieldSet.end() ){
-        const TCHAR* t = *itr;
-        ret[j]=STRDUP_TtoT(t);
-
-        ++j;
-        ++itr;
-    }
-    ret[fieldSet.size()]=NULL;
-    return ret;
-
-  }
+	TermVectorsReader* SegmentReader::getTermVectorsReader() {
+		TermVectorsReader* tvReader = termVectorsLocal.get();
+		if (tvReader == NULL) {
+		  tvReader = termVectorsReaderOrig->clone();
+		  termVectorsLocal.set(tvReader);
+		}
+		return tvReader;
+	}
 
    TermFreqVector* SegmentReader::getTermFreqVector(int32_t docNumber, const TCHAR* field){
-    // Check if this field is invalid or has no stored term vector
-    FieldInfo* fi = fieldInfos->fieldInfo(field);
-    if (fi == NULL || !fi->storeTermVector) 
-       return NULL;
-
-    return termVectorsReader->get(docNumber, field);
+		if ( field != NULL ){
+			// Check if this field is invalid or has no stored term vector
+			FieldInfo* fi = fieldInfos->fieldInfo(field);
+			if (fi == NULL || !fi->storeTermVector || termVectorsReaderOrig == NULL ) 
+				return NULL;
+		}
+		TermVectorsReader* termVectorsReader = getTermVectorsReader();
+		if (termVectorsReader == NULL)
+		  return NULL;
+		return termVectorsReader->get(docNumber, field);
   }
 
-
-   TermFreqVector** SegmentReader::getTermFreqVectors(int32_t docNumber){
+   bool SegmentReader::getTermFreqVectors(int32_t docNumber, Array<TermFreqVector*>& result) {
+    if (termVectorsReaderOrig == NULL)
+      return false;
+    
+    TermVectorsReader* termVectorsReader = getTermVectorsReader();
     if (termVectorsReader == NULL)
-      return NULL;
-
-    return termVectorsReader->get(docNumber);
+      return false;
+    
+    return termVectorsReader->get(docNumber, result);
   }
+
 CL_NS_END
