@@ -38,9 +38,9 @@ public:
   int32_t doc() const;
   bool next();
   bool skipTo(int32_t target);
-  float_t score() const;
-  virtual Explanation* explain(int32_t doc) {
-    _CLTHROWA(CL_ERR_UnsupportedOperation,"UnsupportedOperationException()");
+  float_t score();
+  virtual void explain(int32_t doc, Explanation* ret) {
+    _CLTHROWA(CL_ERR_UnsupportedOperation,"UnsupportedOperationException: ConjunctionScorer::explain");
   }
 
 
