@@ -6,6 +6,7 @@
 ------------------------------------------------------------------------------*/
 #include "CLucene/StdHeader.h"
 #include "FSDirectory.h"
+#include "CLucene/index/IndexReader.h"
 #include "CLucene/util/Misc.h"
 #include "CLucene/util/MD5Digester.h"
 #include "CLucene/debug/condition.h"
@@ -21,6 +22,8 @@ CL_NS_USE(util)
    */
 	static CL_NS(util)::CLHashMap<const char*,FSDirectory*,CL_NS(util)::Compare::Char,CL_NS(util)::Equals::Char> DIRECTORIES(false,false);
 
+	bool FSDirectory::disableLocks=false;
+
 	FSDirectory::FSIndexInput::FSIndexInput(const char* path, int32_t __bufferSize):
 		BufferedIndexInput(__bufferSize)	
     {
@@ -123,6 +126,9 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 		_CLTHROWA(CL_ERR_IO, "read past EOF");
 	}
 	if (bufferLength == -1){
+		//if (EINTR == errno) we could do something else... but we have
+		//to guarantee some return, or throw EOF
+		
 		_CLTHROWA(CL_ERR_IO, "read error");
 	}
 	_pos+=bufferLength;
@@ -153,7 +159,10 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 	if ( fhandle >= 0 ){
 	  try {
         FSIndexOutput::close();
-	  }catch(...){ //todo: only catch IO Err???
+	  }catch(CLuceneError& err){
+	    //ignore IO errors...
+	    if ( err.number() != CL_ERR_IO )
+	        throw;
 	  }
 	}
   }
@@ -166,8 +175,12 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
   }
   void FSDirectory::FSIndexOutput::close() {
     try{
-      IndexOutput::close();
-    }catch(...){} //todo: only catch IO Err???
+      BufferedIndexOutput::close();
+    }catch(CLuceneError& err){
+	    //ignore IO errors...
+	    if ( err.number() != CL_ERR_IO )
+	        throw;
+    }
 
     if ( _close(fhandle) != 0 )
       _CLTHROWA(CL_ERR_IO, "File IO Close error");
@@ -177,7 +190,7 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 
   void FSDirectory::FSIndexOutput::seek(const int64_t pos) {
     CND_PRECONDITION(fhandle>=0,"file is not open");
-    IndexOutput::seek(pos);
+    BufferedIndexOutput::seek(pos);
 	int64_t ret = fileSeek(fhandle,pos,SEEK_SET);
 	if ( ret != pos ){
       _CLTHROWA(CL_ERR_IO, "File IO Seek error");
@@ -192,22 +205,23 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 	const char* FSDirectory::LOCK_DIR=NULL;
 	const char* FSDirectory::getLockDir(){
 		#ifdef LUCENE_LOCK_DIR
-		LOCK_DIR = LUCENE_LOCK_DIR	
+		LOCK_DIR = LUCENE_LOCK_DIR;
 		#else	
-		if ( LOCK_DIR == NULL ){
 			#ifdef LUCENE_LOCK_DIR_ENV_1
+			if ( LOCK_DIR == NULL )
 				LOCK_DIR = getenv(LUCENE_LOCK_DIR_ENV_1);
 			#endif			
 			#ifdef LUCENE_LOCK_DIR_ENV_2
-				if ( LOCK_DIR == NULL && getenv(LUCENE_LOCK_DIR_ENV_2) != NULL )
-					LOCK_DIR = getenv(LUCENE_LOCK_DIR_ENV_2);
+			if ( LOCK_DIR == NULL )
+				LOCK_DIR = getenv(LUCENE_LOCK_DIR_ENV_2);
 			#endif
 			#ifdef LUCENE_LOCK_DIR_ENV_FALLBACK
-				if ( LOCK_DIR == NULL )
-					LOCK_DIR=LUCENE_LOCK_DIR_ENV_FALLBACK;
+			if ( LOCK_DIR == NULL )
+				LOCK_DIR=LUCENE_LOCK_DIR_ENV_FALLBACK;
 			#endif
-		}
-	#endif
+			if ( LOCK_DIR == NULL )
+				_CLTHROWA(CL_ERR_IO, "Couldn't get determine lock dir");
+		#endif
 		
 		return LOCK_DIR;
 	}
@@ -222,12 +236,25 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
   		strcpy(directory,path);	
   	}
     
-	const char* tmplockdir = getLockDir();
+		const char* tmplockdir = getLockDir();
     if ( tmplockdir == NULL ) {
       strcpy(lockDir,directory);
     } else {
       strcpy(lockDir,tmplockdir);
     }
+    
+    // Ensure that lockDir exists and is a directory.
+		struct fileStat fstat;
+	  if ( fileStat(tmplockdir,&fstat) != 0 ) {
+			//todo: should construct directory using _mkdirs... have to write replacement
+			if ( _mkdir(directory) == -1 ){
+				_CLTHROWA(CL_ERR_IO,"Cannot create temp directory"); //todo: make richer error
+			}
+		}
+		if ( !(fstat.st_mode & S_IFDIR) ){
+			_CLTHROWA(CL_ERR_IO, "Found regular file where directory expected"); //todo: make richer error: " + lockDir);
+    }
+	
 
     if (createDir) {
       create();
@@ -244,15 +271,22 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 
   void FSDirectory::create(){
     SCOPED_LOCK_MUTEX(THIS_LOCK)
-	if ( !Misc::dir_Exists(directory) ){
-       //todo: should construct directory using _mkdirs... have to write replacement
+		struct fileStat fstat;
+    if ( fileStat(directory,&fstat) != 0 ) {
+	  	//todo: should construct directory using _mkdirs... have to write replacement
       if ( _mkdir(directory) == -1 ){
-		  char* err = _CL_NEWARRAY(char,27+strlen(directory)+1); //27: len of "Couldn't create directory: "
-		  strcpy(err,"Couldn't create directory: ");
-		  strcat(err,directory);
-		  _CLTHROWA_DEL(CL_ERR_IO, err );
+			  char* err = _CL_NEWARRAY(char,27+strlen(directory)+1); //27: len of "Couldn't create directory: "
+			  strcpy(err,"Couldn't create directory: ");
+			  strcat(err,directory);
+			  _CLTHROWA_DEL(CL_ERR_IO, err );
       }
-	}
+		}
+
+		if ( fileStat(directory,&fstat) != 0 || !(fstat.st_mode & S_IFDIR) ){
+	      char tmp[1024];
+	      _snprintf(tmp,1024,"%s not a directory", directory);
+	      _CLTHROWA(CL_ERR_IO,tmp);
+		}
 
 	  //clear old files
     DIR* dir = opendir(directory);
@@ -261,34 +295,38 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 
     char path[CL_MAX_DIR];
     while ( fl != NULL ){
-      _snprintf(path,CL_MAX_DIR,"%s/%s",directory,fl->d_name);
-      int32_t ret = fileStat(path,&buf);
-      if ( ret==0 && !(buf.st_mode & S_IFDIR) ) {
-        if ( (strcmp(fl->d_name, ".")) && (strcmp(fl->d_name, "..")) ) {
-        if ( _unlink( path ) == -1 ) {
-          closedir(dir);
-          _CLTHROWA(CL_ERR_IO, "Couldn't delete file "); //todo: make richer error
-        }
-        }
-      }
-      fl = readdir(dir);
+		  if ( CL_NS(index)::IndexReader::isLuceneFile(fl->d_name) ){
+				_snprintf(path,CL_MAX_DIR,"%s/%s",directory,fl->d_name);
+				int32_t ret = fileStat(path,&buf);
+				if ( ret==0 && !(buf.st_mode & S_IFDIR) ) {
+					if ( (strcmp(fl->d_name, ".")) && (strcmp(fl->d_name, "..")) ) {
+						if ( _unlink( path ) == -1 ) {
+						  closedir(dir);
+						  _CLTHROWA(CL_ERR_IO, "Couldn't delete file "); //todo: make richer error
+						}
+					}
+				}
+		  }
+		  fl = readdir(dir);
     }
     closedir(dir);
 
-	char* lockPrefix = getLockPrefix(); // clear old locks
-	size_t lockPrefixLen = strlen(lockPrefix);
+		char* lockPrefix = getLockPrefix(); // clear old locks
+		size_t lockPrefixLen = strlen(lockPrefix);
 	
     dir = opendir(lockDir);
-	fl = readdir(dir);
+		if ( dir == NULL )
+			_CLTHROWA(CL_ERR_IO, "Cannot read lock directory"); //todo: richer error: + lockDir.getAbsolutePath());
+		fl = readdir(dir);
     while ( fl != NULL ){
-		if (strncmp(fl->d_name,lockPrefix,lockPrefixLen)==0){
-			_snprintf(path,CL_MAX_DIR,"%s/%s",lockDir,fl->d_name);
-			if ( _unlink( path ) == -1 ) {
-				closedir(dir);
-				_CLDELETE_CaARRAY(lockPrefix);
-				_CLTHROWA(CL_ERR_IO, "Couldn't delete file "); //todo: make richer error
+			if (strncmp(fl->d_name,lockPrefix,lockPrefixLen)==0){
+				_snprintf(path,CL_MAX_DIR,"%s/%s",lockDir,fl->d_name);
+				if ( _unlink( path ) == -1 ) {
+					closedir(dir);
+					_CLDELETE_CaARRAY(lockPrefix);
+					_CLTHROWA(CL_ERR_IO, "Couldn't delete file "); //todo: make richer error
+				}
 			}
-		}
       fl = readdir(dir);
     }
     closedir(dir);
@@ -305,13 +343,12 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
   FSDirectory::~FSDirectory(){
   }
 
-  char** FSDirectory::list() const{
+  void FSDirectory::list(vector<string>* names) const{ //todo: fix this, ugly!!!
     CND_PRECONDITION(directory[0]!=0,"directory is not open");
     DIR* dir = opendir(directory);
     
     struct dirent* fl = readdir(dir);
     struct fileStat buf;
-    AStringArray names;
 
     char path[CL_MAX_DIR];
 	strncpy(path,directory,CL_MAX_DIR);
@@ -322,18 +359,11 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
       strcpy(pathP,fl->d_name);
       fileStat(path,&buf);
       if ( !(buf.st_mode & S_IFDIR) ) {
-        names.push_back( STRDUP_AtoA(fl->d_name) );
+        names->push_back( fl->d_name );
       }
       fl = readdir(dir);
     }
     closedir(dir);
-
-    size_t size = names.size();
-    char** ret = _CL_NEWARRAY(char*,size+1);
-    for ( size_t i=0;i<size;i++ )
-      ret[i] = names[i];
-    ret[size]=NULL;
-    return ret;
   }
 
   bool FSDirectory::fileExists(const char* name) const {
@@ -397,7 +427,6 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
     char buffer[CL_MAX_DIR];
     _snprintf(buffer,CL_MAX_DIR,"%s%s%s",directory,PATH_DELIMITERA,name);
 	
-	//todo: haven't checked this:
     int32_t r = _open(buffer, O_RDWR, _S_IWRITE);
 	if ( r < 0 )
 		_CLTHROWA(CL_ERR_IO,"IO Error while touching file");
@@ -416,7 +445,7 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
   }
 
   IndexInput* FSDirectory::openInput(const char* name ) {
-  	return openInput(name, LUCENE_STREAM_BUFFER_SIZE);
+  	return openInput(name, CL_NS(store)::BufferedIndexOutput::BUFFER_SIZE);
   }
   
 #ifdef LUCENE_FS_MMAP
@@ -446,20 +475,20 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
   }
 		
   void FSDirectory::close(){
-      SCOPED_LOCK_MUTEX(DIRECTORIES.THIS_LOCK)
-      {
-	      SCOPED_LOCK_MUTEX(THIS_LOCK)
-
-	      CND_PRECONDITION(directory[0]!=0,"directory is not open");
-
-          if (--refCount <= 0 ) {//refcount starts at 1
-              Directory* dir = DIRECTORIES.get(getDirName());
-              if(dir){
-                DIRECTORIES.remove( getDirName() ); //this will be removed in ~FSDirectory
-			    _CLDECDELETE(dir);
-              }
-	       }
-       }
+    SCOPED_LOCK_MUTEX(DIRECTORIES.THIS_LOCK)
+    {
+	    SCOPED_LOCK_MUTEX(THIS_LOCK)
+	
+	    CND_PRECONDITION(directory[0]!=0,"directory is not open");
+	
+	    if (--refCount <= 0 ) {//refcount starts at 1
+	        Directory* dir = DIRECTORIES.get(getDirName());
+	        if(dir){
+	            DIRECTORIES.remove( getDirName() ); //this will be removed in ~FSDirectory
+	            _CLDECDELETE(dir);
+	        }
+	    }
+	}
    }
 
    /**
@@ -489,15 +518,15 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 	    return ret; 
   }
 
-  bool FSDirectory::deleteFile(const char* name)  {
+  bool FSDirectory::doDeleteFile(const char* name)  {
 	CND_PRECONDITION(directory[0]!=0,"directory is not open");
     char fl[CL_MAX_DIR];
     priv_getFN(fl, name);
 	return _unlink(fl) != -1;
   }
-
+  
   void FSDirectory::renameFile(const char* from, const char* to){
-	  CND_PRECONDITION(directory[0]!=0,"directory is not open");
+	CND_PRECONDITION(directory[0]!=0,"directory is not open");
     SCOPED_LOCK_MUTEX(THIS_LOCK)
     char old[CL_MAX_DIR];
     priv_getFN(old, from);
@@ -509,22 +538,35 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
     delete() and the call to renameTo() then we're screwed, but I've
     been unable to figure out how else to do this... */
 
-    if ( Misc::dir_Exists(nu) )
-      if( _unlink(nu) != 0 ){
-	    char* err = _CL_NEWARRAY(char,16+strlen(to)+1); //16: len of "couldn't delete "
-		strcpy(err,"couldn't delete ");
-		strcat(err,to);
-        _CLTHROWA_DEL(CL_ERR_IO, err );
+    if ( Misc::dir_Exists(nu) ){
+      //we run this sequence of unlinking an arbitary 100 times
+      //on some platforms (namely windows), there can be a
+      //delay between unlink and dir_exists==false          
+      while ( true ){
+          if( _unlink(nu) != 0 ){
+    	    char* err = _CL_NEWARRAY(char,16+strlen(to)+1); //16: len of "couldn't delete "
+    		strcpy(err,"couldn't delete ");
+    		strcat(err,to);
+            _CLTHROWA_DEL(CL_ERR_IO, err );
+          }
+          //we can wait until the dir_Exists() returns false
+          //after the success run of unlink()
+          int i=0;
+          while ( Misc::dir_Exists(nu) && i < 100 )
+            i++;
+          if ( !Misc::dir_Exists(nu) )
+            break; //keep trying to unlink until the file is gone, or the unlink fails.
       }
+    }
     if ( _rename(old,nu) != 0 ){
        //todo: jlucene has some extra rename code - if the rename fails, it copies
        //the whole file to the new file... might want to implement that if renaming
        //fails on some platforms
-        char buffer[200];
+        char buffer[20+CL_MAX_PATH+CL_MAX_PATH];
         strcpy(buffer,"couldn't rename ");
         strcat(buffer,from);
         strcat(buffer," to ");
-        strcat(buffer,to);
+        strcat(buffer,nu);
       _CLTHROWA(CL_ERR_IO, buffer );
     }
   }
@@ -533,6 +575,14 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 	CND_PRECONDITION(directory[0]!=0,"directory is not open");
     char fl[CL_MAX_DIR];
     priv_getFN(fl, name);
+	if ( Misc::dir_Exists(fl) ){
+		if ( _unlink(fl) != 0 ){
+			char tmp[1024];
+			strcpy(tmp, "Cannot overwrite: ");
+			strcat(tmp, name);
+			_CLTHROWA(CL_ERR_IO, tmp);
+		}
+	}
     return _CLNEW FSIndexOutput( fl );
   }
 
@@ -572,12 +622,9 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 	  return ret;
   }
   bool FSDirectory::FSLock::obtain() {
-   if (LUCENE_DISABLE_LOCKS)
+   if (disableLocks)
       return true;
 
-	if ( Misc::dir_Exists(lockFile) )
-	  return false;
-
 	if ( !Misc::dir_Exists(lockDir) ){
        //todo: should construct directory using _mkdirs... have to write replacement
       if ( _mkdir(lockDir) == -1 ){
@@ -587,7 +634,8 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 		  _CLTHROWA_DEL(CL_ERR_IO, err );
       }
     }
-    int32_t r = _open(lockFile,  O_RDWR | O_CREAT | O_RANDOM , _S_IREAD | _S_IWRITE); //must do this or file will be created Read only
+    int32_t r = _open(lockFile,  O_RDWR | O_CREAT | O_RANDOM | O_EXCL, 
+    	_S_IREAD | _S_IWRITE); //must do this or file will be created Read only
 	if ( r < 0 )
 	  return false;
 	else{
@@ -597,13 +645,13 @@ void FSDirectory::FSIndexInput::readInternal(uint8_t* b, const int32_t len) {
 
   }
   bool FSDirectory::FSLock::isLocked(){
-     if (LUCENE_DISABLE_LOCKS)
+     if (disableLocks)
           return false;
 
      return Misc::dir_Exists(lockFile);
   }
   void FSDirectory::FSLock::release() {
-    if (LUCENE_DISABLE_LOCKS)
+    if (disableLocks)
           return;
     _unlink( lockFile );
   }
