@@ -62,7 +62,6 @@ Searcher* sort_getIndex (bool even, bool odd){
 			writer.addDocument (&doc);
 		}
 	}
-	writer.optimize ();
 	writer.close ();
 	IndexSearcher* res = _CLNEW IndexSearcher(indexStore);
 	_CLDECDELETE(indexStore);
@@ -336,11 +335,11 @@ void testReverseSort(CuTest *tc){
 	/*SortField* sorts[3] = { _CLNEW SortField (NULL, SortField::INT,true), SortField::FIELD_DOC, NULL };
 	_sort->setSort (sorts);
 	sortMatches (tc, sort_full, sort_queryX, _sort, _T("IEGCA"));
-	sortMatches (tc, sort_full, sort_queryY, _sort, _T("JFHDB"));
+	sortMatches (tc, sort_full, sort_queryY, _sort, _T("JFHDB"));*/
 
 	_sort->setSort (_CLNEW SortField (NULL, SortField::DOC, true));
 	sortMatches (tc, sort_full, sort_queryX, _sort, _T("IGECA"));
-	sortMatches (tc, sort_full, sort_queryY, _sort, _T("JHFDB"));*/
+	sortMatches (tc, sort_full, sort_queryY, _sort, _T("JHFDB"));
 
 	_sort->setSort (_T("int"), true);
 	sortMatches (tc, sort_full, sort_queryX, _sort, _T("CAEGI"));
